"""This module contains forms used for admin integration."""

from copy import deepcopy

from django.contrib.admin.widgets import AdminSplitDateTime
from django.forms import (BooleanField, CharField, ChoiceField, MultipleChoiceField, DateTimeField,
                          FloatField, IntegerField, ModelForm)
from django.utils.translation import ugettext_lazy as _


class BaseDynamicEntityForm(ModelForm):
    """
    ``ModelForm`` for entity with support for EAV attributes. Form fields are
    created on the fly depending on schema defined for given entity instance.
    If no schema is defined (i.e. the entity instance has not been saved yet),
    only static fields are used. However, on form validation the schema will be
    retrieved and EAV fields dynamically added to the form, so when the
    validation is actually done, all EAV fields are present in it (unless
    Rubric is not defined).

    Mapping between attribute types and field classes is as follows:

    =====  =============
    Type      Field
    =====  =============
    text   CharField
    float  IntegerField
    int    DateTimeField
    bool   BooleanField
    enum   ChoiceField
    =====  =============
    """
    FIELD_CLASSES = {
        'text': CharField,
        'float': FloatField,
        'int': IntegerField,
        'date': DateTimeField,
        'bool': BooleanField,
        'enum': ChoiceField,
        'm2m_enum': MultipleChoiceField
    }

    def __init__(self, data=None, *args, **kwargs):
        super(BaseDynamicEntityForm, self).__init__(data, *args, **kwargs)
        config_cls = self.instance._eav_config_cls
        self.entity = getattr(self.instance, config_cls.eav_attr)
        self._build_dynamic_fields()

    def _build_dynamic_fields(self):
        # Reset form fields.
        self.fields = deepcopy(self.base_fields)

        for attribute in self.entity.get_all_attributes():
            value = getattr(self.entity, attribute.slug)

            defaults = {
                'label': attribute.name.capitalize(),
                'required': attribute.required,
                'help_text': attribute.help_text,
                'validators': attribute.get_validators(),
            }

            datatype = attribute.datatype

            if datatype in [attribute.TYPE_ENUM, attribute.TYPE_M2M]:
                values = attribute.get_choices().values_list('id', 'value')
                choices = [('', '-----')] + list(values)
                defaults.update({'choices': choices})

                if value:
                    if datatype == attribute.TYPE_ENUM:
                        defaults.update({'initial': value.pk})
                    else:
                        defaults.update({'initial': [_ for _ in value.values_list('pk', flat=True)]})

            elif datatype == attribute.TYPE_DATE:
                defaults.update({'widget': AdminSplitDateTime})
            elif datatype == attribute.TYPE_OBJECT:
                continue

            MappedField = self.FIELD_CLASSES[datatype]
            self.fields[attribute.slug] = MappedField(**defaults)

            # Fill initial data (if attribute was already defined).
            if value and not datatype in [attribute.TYPE_ENUM, attribute.TYPE_M2M]:
                self.initial[attribute.slug] = value

    def save(self, commit=True):
        """
        Saves this ``form``'s cleaned_data into model instance
        ``self.instance`` and related EAV attributes. Returns ``instance``.
        """
        if self.errors:
            raise ValueError(_(
                'The %s could not be saved because the data'
                'didn\'t validate.' % self.instance._meta.object_name
            ))

        # Create entity instance, don't save yet.
        instance = super(BaseDynamicEntityForm, self).save(commit=False)

        # Assign attributes.
        for attribute in self.entity.get_all_attributes():
            value = self.cleaned_data.get(attribute.slug)

            if attribute.datatype == attribute.TYPE_ENUM:
                if value:
                    value = attribute.enum_group.values.get(pk=value)
                else:
                    value = None

            if attribute.datatype == attribute.TYPE_M2M:
                if value:
                    values = attribute.enum_group.values.filter(pk=value)
                else:
                    values = []

                m2m_attribute = getattr(self.entity, attribute.slug)
                m2m_attribute.clear()
                m2m_attribute.add(**values)

            setattr(self.entity, attribute.slug, value)

        # Save entity and its attributes.
        if commit:
            instance.save()
            self._save_m2m()

        return instance



# Generated by Django 3.1.1 on 2020-11-11 11:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('eav', '0005_auto_20201104_0531'),
    ]

    operations = [
        migrations.AlterField(
            model_name='value',
            name='generic_value_ct',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='value_values', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_enum',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='eav_values', to='eav.enumvalue'),
        ),
    ]



"""
This module contains a validator for each :class:`~eav.models.Attribute` datatype.

A validator is a callable that takes a value and raises a ``ValidationError``
if it doesn't meet some criteria (see `Django validators
<http://docs.djangoproject.com/en/dev/ref/validators/>`_).

These validators are called by the
:meth:`~eav.models.Attribute.validate_value` method in the
:class:`~eav.models.Attribute` model.
"""

import datetime

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _


def validate_text(value):
    """
    Raises ``ValidationError`` unless *value* type is ``str`` or ``unicode``
    """
    if not isinstance(value, str):
        raise ValidationError(_(u"Must be str or unicode"))


def validate_float(value):
    """
    Raises ``ValidationError`` unless *value* can be cast as a ``float``
    """
    try:
        float(value)
    except ValueError:
        raise ValidationError(_(u"Must be a float"))


def validate_int(value):
    """
    Raises ``ValidationError`` unless *value* can be cast as an ``int``
    """
    try:
        int(value)
    except ValueError:
        raise ValidationError(_(u"Must be an integer"))


def validate_date(value):
    """
    Raises ``ValidationError`` unless *value* is an instance of ``datetime``
    or ``date``
    """
    if not isinstance(value, datetime.datetime) and not isinstance(value, datetime.date):
        raise ValidationError(_(u"Must be a date or datetime"))


def validate_bool(value):
    """
    Raises ``ValidationError`` unless *value* type is ``bool``
    """
    if not isinstance(value, bool):
        raise ValidationError(_(u"Must be a boolean"))


def validate_object(value):
    """
    Raises ``ValidationError`` unless *value* is a saved
    django model instance.
    """
    if not isinstance(value, models.Model):
        raise ValidationError(_(u"Must be a django model object instance"))

    if not value.pk:
        raise ValidationError(_(u"Model has not been saved yet"))


def validate_enum(value):
    """
    Raises ``ValidationError`` unless *value* is a saved
    :class:`~eav.models.EnumValue` model instance.
    """
    from eav.models import EnumValue

    if isinstance(value, EnumValue) and not value.pk:
        raise ValidationError(_(u"EnumValue has not been saved yet"))


def validate_m2m_enum(value):
    pass



# Generated by Django 3.1.1 on 2020-10-02 14:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eav', '0003_attribute_category'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='attribute',
            name='category',
        ),
        migrations.AddField(
            model_name='attribute',
            name='tag',
            field=models.CharField(blank=True, help_text='Tag the attribute for categorisation', max_length=128, null=True, verbose_name='Tag'),
        ),
    ]



# -*- coding: utf-8 -*-
"""
This module contains custom :class:`EavQuerySet` class used for overriding
relational operators and pure functions for rewriting Q-expressions.
Q-expressions need to be rewritten for two reasons:

1. In order to hide implementation from the user and provide easy to use
   syntax sugar, i.e.::

       Supplier.objects.filter(eav__city__startswith='New')

   instead of::

       city_values = Value.objects.filter(value__text__startswith='New')
       Supplier.objects.filter(eav_values__in=city_values)

   For details see: :func:`eav_filter`.

2. To ensure that Q-expression tree is compiled to valid SQL.
   For details see: :func:`rewrite_q_expr`.
"""
from itertools import count
from functools import wraps

from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import FieldError, ObjectDoesNotExist
from django.db import models
from django.db.models import Case, IntegerField, Q, When, Count
from django.db.models.query import QuerySet
from django.db.utils import NotSupportedError

from eav.models import Attribute, Value, EnumValue


def is_eav_and_leaf(expr, gr_name):
    """
    Checks whether Q-expression is an EAV AND leaf.

    Args:
        expr (Union[Q, tuple]): Q-expression to be checked.
        gr_name (str): Generic relation attribute name, by default 'eav_values'

    Returns:
        bool
    """
    return (
            getattr(expr, 'connector', None) == 'AND' and
            len(expr.children) == 1 and
            expr.children[0][0] in ['pk__in', '{}__in'.format(gr_name)]
    )


def rewrite_q_expr(model_cls, expr):
    """
    Rewrites Q-expression to safe form, in order to ensure that
    generated SQL is valid.

IGNORE:
    Suppose we have the following Q-expression:

    └── OR
         ├── AND
         │    └── eav_values__in [1, 2, 3]
         └── AND (1)
              ├── AND
              │    └── eav_values__in [4, 5]
              └── AND
                   └── eav_values__in [6, 7, 8]
IGNORE

    All EAV values are stored in a single table. Therefore, INNER JOIN
    generated for the AND-expression (1) will always fail, i.e.
    single row in a eav_values table cannot be both in two disjoint sets at
    the same time (and the whole point of using AND, usually, is two have
    two different sets). Therefore, we must paritially rewrite the
    expression so that the generated SQL is valid::

IGNORE:
    └── OR
         ├── AND
         │    └── eav_values__in [1, 2, 3]
         └── AND
              └── pk__in [1, 2]
IGNORE

    This is done by merging dangerous AND's and substituting them with
    explicit ``pk__in`` filter, where pks are taken from evaluted
    Q-expr branch.

    Args:
        model_cls (TypeVar): model class used to construct :meth:`QuerySet`
            from leaf attribute-value expression.
            expr: (Q | tuple): Q-expression (or attr-val leaf) to be rewritten.

    Returns:
        Union[Q, tuple]
    """
    # Node in a Q-expr can be a Q or an attribute-value tuple (leaf).
    # We are only interested in Qs.

    if isinstance(expr, Q):
        config_cls = getattr(model_cls, '_eav_config_cls', None)
        gr_name = config_cls.generic_relation_attr

        # Recurively check child nodes.
        expr.children = [rewrite_q_expr(model_cls, c) for c in expr.children]
        # Check which ones need a rewrite.
        rewritable = [c for c in expr.children if is_eav_and_leaf(c, gr_name)]

        # Conflict occurs only with two or more AND-expressions.
        # If there is only one we can ignore it.

        if len(rewritable) > 1:
            q = None
            # Save nodes which shouldn't be merged (non-EAV).
            other = [c for c in expr.children if not c in rewritable]

            for child in rewritable:
                if not (child.children and len(child.children) == 1):
                    raise AssertionError('Child must have exactly one descendant')
                # Child to be merged is always a terminal Q node,
                # i.e. it's an AND expression with attribute-value tuple child.
                attrval = child.children[0]
                if not isinstance(attrval, tuple):
                    raise AssertionError('Attribute-value must be a tuple')

                fname = '{}__in'.format(gr_name)

                # Child can be either a 'eav_values__in' or 'pk__in' query.
                # If it's the former then transform it into the latter.
                # Otherwise, check if the value is valid for the '__in' query.
                # If so, reverse it back to QuerySet so that set operators
                # can be applied.

                if attrval[0] == fname or hasattr(attrval[1], '__contains__'):
                    # Create model queryset.
                    _q = model_cls.objects.filter(**{fname: attrval[1]})
                else:
                    # Second val in tuple is a queryset.
                    _q = attrval[1]

                # Explicitly check for None. 'or' doesn't work here
                # as empty QuerySet, which is valid, is falsy.
                q = q if q != None else _q

                if expr.connector == 'AND':
                    q &= _q
                else:
                    q |= _q

            # If any two children were merged,
            # update parent expression.
            if q != None:
                expr.children = other + [('pk__in', q)]

    return expr


def eav_filter(func):
    """
    Decorator used to wrap filter and exclude methods. Passes args through
    :func:`expand_q_filters` and kwargs through :func:`expand_eav_filter`. Returns the
    called function (filter or exclude).
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        nargs = []
        nkwargs = {}

        for arg in args:
            if isinstance(arg, Q):
                # Modify Q objects (warning: recursion ahead).
                arg = expand_q_filters(arg, self.model)
                # Rewrite Q-expression to safeform.
                arg = rewrite_q_expr(self.model, arg)
            nargs.append(arg)

        for key, value in kwargs.items():
            # Modify kwargs (warning: recursion ahead).
            nkey, nval = expand_eav_filter(self.model, key, value)

            if nkey in nkwargs:
                # Apply AND to both querysets.
                nkwargs[nkey] = (nkwargs[nkey] & nval).distinct()
            else:
                nkwargs.update({nkey: nval})

        return func(self, *nargs, **nkwargs)

    return wrapper


def expand_q_filters(q, root_cls):
    """
    Takes a Q object and a model class.
    Recursively passes each filter / value in the Q object tree leaf nodes
    through :func:`expand_eav_filter`.
    """
    new_children = []

    for qi in q.children:
        if isinstance(qi, tuple):
            # This child is a leaf node: in Q this is a 2-tuple of:
            # (filter parameter, value).
            key, value = expand_eav_filter(root_cls, *qi)
            new_children.append(Q(**{key: value}))
        else:
            # This child is another Q node: recursify!
            new_children.append(expand_q_filters(qi, root_cls))

    q.children = new_children
    return q


def expand_eav_filter(model_cls, key, value):
    """
    Accepts a model class and a key, value.
    Recurisively replaces any eav filter with a subquery.

    For example::

        key = 'eav__height'
        value = 5

    Would return::

        key = 'eav_values__in'
        value = Values.objects.filter(value_int=5, attribute__slug='height')
    """
    fields = key.split('__')
    config_cls = getattr(model_cls, '_eav_config_cls', None)

    if len(fields) > 1 and config_cls and fields[0] == config_cls.eav_attr:
        slug = fields[1]
        gr_name = config_cls.generic_relation_attr

        attribute_kwargs = {
            'slug': slug
        }

        try:
            attribute_kwargs['tag'] = config_cls.ATTRIBUTE_TAG
        except AttributeError:
            pass

        datatype = Attribute.objects.get(**attribute_kwargs).datatype

        value_key = ''
        if datatype == Attribute.TYPE_ENUM and not isinstance(value, EnumValue):
            lookup = '__value__{}'.format(fields[2]) if len(fields) > 2 else '__value'
            value_key = 'value_{}{}'.format(datatype, lookup)
        elif datatype == Attribute.TYPE_OBJECT:
            value_key = 'generic_value_id'
        else:
            lookup = '__{}'.format(fields[2]) if len(fields) > 2 else ''
            value_key = 'value_{}{}'.format(datatype, lookup)
        kwargs = {
            value_key: value,
            'attribute__slug': slug
        }
        value = Value.objects.filter(**kwargs)

        return '%s__in' % gr_name, value

    try:
        field = model_cls._meta.get_field(fields[0])
    except FieldDoesNotExist:
        return key, value

    if not field.auto_created or field.concrete:
        return key, value
    else:
        sub_key = '__'.join(fields[1:])
        key, value = expand_eav_filter(field.model, sub_key, value)
        return '{}__{}'.format(fields[0], key), value


class EavQuerySet(QuerySet):
    """
    Overrides relational operators for EAV models.
    """

    def enum_exact_filter(self, attribute, enums):

        filter_condition = Q(count=len(enums))

        if not isinstance(attribute, Attribute):

            config_cls = self.model._eav_config_cls

            attribute_kwargs = {
                'slug': attribute
            }

            try:
                attribute_kwargs['tag'] = config_cls.ATTRIBUTE_TAG
            except AttributeError:
                pass

            attribute = Attribute.objects.get(**attribute_kwargs)

        for enum in enums:
            if not isinstance(enum, EnumValue):
                try:
                    enum = attribute.enum_group.values.get(value=enum)
                except Exception as e:
                    raise Exception(
                        "Root Exception: {exc}"
                        "Unique EnumValue with value `{enum_value}` not found for Attribute: `{attribute}`\n"
                        "Try passing list of enum objects instead".format(
                            exc=str(e),
                            enum_value=enum,
                            attribute=attribute.name
                        )
                    )

            attribute_condition = {
                'eav__{}'.format(attribute.slug): enum
            }

            filter_condition = filter_condition & Q(**attribute_condition)

        return super() \
            .annotate(count=Count('eav_values__value_m2m_enum')) \
            .filter(Q(filter_condition)).distinct()

    @eav_filter
    def filter(self, *args, **kwargs):
        """
        Pass *args* and *kwargs* through :func:`eav_filter`, then pass to
        the ``Manager`` filter method.
        """
        return super(EavQuerySet, self).filter(*args, **kwargs)

    @eav_filter
    def exclude(self, *args, **kwargs):
        """
        Pass *args* and *kwargs* through :func:`eav_filter`, then pass to
        the ``Manager`` exclude method.
        """
        return super(EavQuerySet, self).exclude(*args, **kwargs)

    @eav_filter
    def get(self, *args, **kwargs):
        """
        Pass *args* and *kwargs* through :func:`eav_filter`, then pass to
        the ``Manager`` get method.
        """
        return super(EavQuerySet, self).get(*args, **kwargs)

    def order_by(self, *fields):
        # Django only allows to order querysets by direct fields and
        # foreign-key chains. In order to bypass this behaviour and order
        # by EAV attributes, it is required to construct custom order-by
        # clause manually using Django's conditional expressions.
        # This will be slow, of course.
        order_clauses = []
        query_clause = self
        config_cls = self.model._eav_config_cls

        for term in [t.split('__') for t in fields]:
            # Continue only for EAV attributes.
            if len(term) == 2 and term[0] in [config_cls.eav_attr, '-{}'.format(config_cls.eav_attr)] :
                # Retrieve Attribute over which the ordering is performed.

                field_prepend = ''
                if '-' == term[0][0]:
                    field_prepend = '-'
                    term[0] = term[0][1:]

                try:
                    attribute_kwargs = {
                        'slug': term[1],
                    }

                    try:
                        attribute_kwargs['tag'] = config_cls.ATTRIBUTE_TAG
                    except AttributeError:
                        pass

                    attr = Attribute.objects.get(**attribute_kwargs)
                except ObjectDoesNotExist:
                    raise ObjectDoesNotExist(
                        'Cannot find EAV attribute "{}"'.format(term[1])
                    )

                field_name = 'value_%s' % attr.datatype

                pks_values = Value.objects.filter(
                    # Retrieve pk-values pairs of the related values
                    # (i.e. values for the specified attribute and
                    # belonging to entities in the queryset).
                    attribute__slug=attr.slug,
                    entity_id__in=self
                ).order_by(
                    # Order values by their value-field of
                    # appropriate attribute data-type.
                    '{}{}'.format(field_prepend, field_name)
                ).values_list(
                    # Retrieve only primary-keys of the entities
                    # in the current queryset.
                    'entity_id', field_name
                )

                # Retrive ordered values from pk-value list.
                _, ordered_values = zip(*pks_values)

                # Add explicit ordering and turn
                # list of pairs into look-up table.
                val2ind = dict(zip(ordered_values, count()))

                # Finally, zip ordered pks with their grouped orderings.
                entities_pk = [(pk, val2ind[val]) for pk, val in pks_values]

                # Using ordered primary-keys, construct
                # CASE clause of the form:
                #
                #     CASE
                #         WHEN id = 2 THEN 1
                #         WHEN id = 5 THEN 2
                #         WHEN id = 9 THEN 2
                #         WHEN id = 4 THEN 3
                #     END
                #
                when_clauses = [
                    When(id=pk, then=i)
                    for pk, i in entities_pk
                ]

                order_clause = Case(
                    *when_clauses,
                    output_field=IntegerField()
                )

                clause_name = '__'.join(term)
                # Use when-clause to construct
                # custom order-by clause.
                query_clause = query_clause.annotate(
                    **{clause_name: order_clause}
                )

                order_clauses.append(clause_name)

            elif len(term) >= 2 and term[0] in [config_cls.eav_attr, '-{}'.format(config_cls.eav_attr)]:
                raise NotSupportedError(
                    'EAV does not support ordering through '
                    'foreign-key chains'
                )

            else:
                order_clauses.append(term[0])

        return QuerySet.order_by(query_clause, *order_clauses)



__version__ = '0.13.921'

def register(model_cls, config_cls=None):
    from .registry import Registry
    Registry.register(model_cls, config_cls)

def unregister(model_cls):
    from .registry import Registry
    Registry.unregister(model_cls)



# Generated by Django 2.0.4 on 2018-06-01 09:36

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import eav.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='User-friendly attribute name', max_length=100, verbose_name='Name')),
                ('slug', eav.fields.EavSlugField(help_text='Short unique attribute label', max_length=256, unique=True, verbose_name='Slug')),
                ('description', models.CharField(blank=True, help_text='Short description', max_length=256, null=True, verbose_name='Description')),
                ('datatype', eav.fields.EavDatatypeField(choices=[('text', 'Text'), ('date', 'Date'), ('float', 'Float'), ('int', 'Integer'), ('bool', 'True / False'), ('object', 'Django Object'), ('enum', 'Multiple Choice')], max_length=6, verbose_name='Data Type')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('required', models.BooleanField(default=False, verbose_name='Required')),
                ('display_order', models.PositiveIntegerField(default=1, verbose_name='Display order')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='EnumGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='Name')),
            ],
        ),
        migrations.CreateModel(
            name='EnumValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=50, unique=True, verbose_name='Value')),
            ],
        ),
        migrations.CreateModel(
            name='Value',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entity_id', models.IntegerField()),
                ('value_text', models.TextField(blank=True, null=True)),
                ('value_float', models.FloatField(blank=True, null=True)),
                ('value_int', models.IntegerField(blank=True, null=True)),
                ('value_date', models.DateTimeField(blank=True, null=True)),
                ('value_bool', models.NullBooleanField()),
                ('generic_value_id', models.IntegerField(blank=True, null=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('attribute', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='eav.Attribute', verbose_name='Attribute')),
                ('entity_ct', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='value_entities', to='contenttypes.ContentType')),
                ('generic_value_ct', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='value_values', to='contenttypes.ContentType')),
                ('value_enum', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='eav_values', to='eav.EnumValue')),
            ],
        ),
        migrations.AddField(
            model_name='enumgroup',
            name='values',
            field=models.ManyToManyField(to='eav.EnumValue', verbose_name='Enum group'),
        ),
        migrations.AddField(
            model_name='attribute',
            name='enum_group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='eav.EnumGroup', verbose_name='Choice Group'),
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-01 04:02

from django.db import migrations, models
import eav.fields


class Migration(migrations.Migration):

    dependencies = [
        ('eav', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='value',
            name='value_m2m_enum',
            field=models.ManyToManyField(related_name='eav_m2m_values', to='eav.EnumValue'),
        ),
        migrations.AlterField(
            model_name='attribute',
            name='datatype',
            field=eav.fields.EavDatatypeField(choices=[('text', 'Text'), ('date', 'Date'), ('float', 'Float'), ('int', 'Integer'), ('bool', 'True / False'), ('object', 'Django Object'), ('enum', 'Single Choice'), ('m2m_enum', 'Multiple Choice')], max_length=16, verbose_name='Data Type'),
        ),
    ]



# Generated by Django 3.1.1 on 2020-11-04 05:31

from django.db import migrations
import eav.fields


class Migration(migrations.Migration):

    dependencies = [
        ('eav', '0004_auto_20201002_1407'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attribute',
            name='slug',
            field=eav.fields.EavSlugField(help_text='Short unique attribute label', max_length=256, verbose_name='Slug'),
        ),
        migrations.AlterUniqueTogether(
            name='attribute',
            unique_together={('slug', 'tag')},
        ),
    ]



from setuptools import setup, find_packages

setup(
    name             = 'travisb-django-eav2',
    version          = __import__('eav').__version__,
    license          = 'GNU Lesser General Public License (LGPL), Version 3',
    requires         = ['python (>= 3.5)', 'django (>= 1.11.14)'],
    provides         = ['eav'],
    description      = 'Entity-Attribute-Value storage for Django. Fork of django-eav-2 by Iwo Herka [hi@iwoherka.eu]',
    url              = 'http://github.com/sree93/django-eav2',
    packages         = find_packages(),
    maintainer       = 'Srinath Menon',
    maintainer_email = 'srinath@travisbickle.tech',

    classifiers  = [
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Programming Language :: Python',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import re

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _


class EavSlugField(models.SlugField):
    """
    The slug field used by :class:`~eav.models.Attribute`
    """

    def validate(self, value, instance):
        """
        Slugs are used to convert the Python attribute name to a database
        lookup and vice versa. We need it to be a valid Python identifier. We
        don't want it to start with a '_', underscore will be used in
        variables we don't want to be saved in the database.
        """
        super(EavSlugField, self).validate(value, instance)
        slug_regex = r'[a-z][a-z0-9_]*'

        if not re.match(slug_regex, value):
            raise ValidationError(_(
                'Must be all lower case, start with a letter, and contain '
                'only letters, numbers, or underscores.'
            ))

    @staticmethod
    def create_slug_from_name(name):
        """Creates a slug based on the name."""
        name = name.strip().lower()

        # Change spaces to underscores.
        name = '_'.join(name.split())

        # Remove non alphanumeric characters.
        return re.sub('[^\w]', '', name)


class EavDatatypeField(models.CharField):
    """
    The datatype field used by :class:`~eav.models.Attribute`.
    """

    def validate(self, value, instance):
        """
        Raise ``ValidationError`` if they try to change the datatype of an
        :class:`~eav.models.Attribute` that is already used by
        :class:`~eav.models.Value` objects.
        """
        super(EavDatatypeField, self).validate(value, instance)

        if not instance.pk:
            return

        if type(instance).objects.get(pk=instance.pk).datatype == instance.datatype:
            return

        if instance.value_set.count():
            raise ValidationError(_(
                'You cannot change the datatype of an attribute that is already in use.'
            ))
